#! c:/Python26/python.exe

import os
import re
import web
import copy
import time
import heapq
import bisect
import hashlib
import tempita
import calendar
import datetime

from web_helper import *
from model import *

# =====================================
# Variables & Constant
# =====================================

APPLICATION_URL = None
GRAPHICS_URL = None
DATA_FILE_NAME = None

urls = (
    '/', 'index',
    '/daily/{0,1}', 'daily',
    '/weekly/{0,1}', 'weekly',
    '/monthly/{0,1}', 'monthly',
    '/all/{0,1}', 'all',

    '/del/(.*)/(.*)', 'delete',
    '/done/(.*)/(.*)', 'done',
    '/move/(.*)/(.*)/(.*)', 'move',
    '/repeat/(.*)/(.*)', 'repeat',
    '/do_today/(.*)/(.*)', 'do_today',

    '/config/{0,1}', 'config',
    '/help/{0,1}', 'help',
    '/import/{0,1}', 'import_json',
    '/excel/{0,1}', 'excel_io',
)


# =====================================
# Controllers
# =====================================


def adv(content):
    if os.path.exists("adv.txt"):
        return content.replace('<span class="adv"/>', file("adv.txt").read())
    return content


def is_demo():
    return os.path.exists("demo.txt")


class help:
    def GET(self):
        if not config.is_initialized():
            return
        global APPLICATION_URL, GRAPHICS_URL
        application_url = APPLICATION_URL
        graphics_url = GRAPHICS_URL
        web.header("Content-Type", "text/html; charset=UTF-8")
        template = tempita.HTMLTemplate.from_filename("help.html")
        print adv(template.substitute(locals()))


def show_page(type, data_file=None):
    type = type.split("_")[0]
    pages = {"all":show_all_page,
             "today":show_today_page,
             "daily":show_daily_review_page,
             "weekly":show_weekly_review_page,
             "monthly":show_monthly_review_page}
    page_view_function = pages.get(type, show_today_page)
    page_view_function(data_file)


def show_today_page(data_file=None):
    if not config.is_initialized():
        return
    global APPLICATION_URL, GRAPHICS_URL
    application_url = APPLICATION_URL
    graphics_url = GRAPHICS_URL

    web.header("Content-Type", "text/html; charset=utf-8")

    if data_file is None:
        data_file = LockablePickleFile(DATA_FILE_NAME)
        data_file.read(default=TaskList())

    unfinished_task, today = data_file.content.classified_task("today")
    memo = data_file.content.memo
    free_task = data_file.content.free_task
    template = tempita.HTMLTemplate.from_filename("today.html")
    print adv(template.substitute(locals()))


def show_daily_review_page(data_file=None):
    if not config.is_initialized():
        return
    global APPLICATION_URL, GRAPHICS_URL
    application_url = APPLICATION_URL
    graphics_url = GRAPHICS_URL

    web.header("Content-Type", "text/html; charset=utf-8")

    if data_file is None:
        data_file = LockablePickleFile(DATA_FILE_NAME)
        data_file.read(default=TaskList())

    unfinished_task, this_week, today = data_file.content.classified_task("daily_review")
    memo = data_file.content.memo
    free_task = data_file.content.free_task
    template = tempita.HTMLTemplate.from_filename("daily.html")
    print adv(template.substitute(locals()))


def show_weekly_review_page(data_file=None):
    if not config.is_initialized():
        return
    global APPLICATION_URL, GRAPHICS_URL
    application_url = APPLICATION_URL
    graphics_url = GRAPHICS_URL

    web.header("Content-Type", "text/html; charset=utf-8")

    if data_file is None:
        data_file = LockablePickleFile(DATA_FILE_NAME)
        data_file.read(default=TaskList())

    unfinished_task, this_month, this_week = data_file.content.classified_task("weekly_review")
    memo = data_file.content.memo
    free_task = data_file.content.free_task
    template = tempita.HTMLTemplate.from_filename("weekly.html")
    print adv(template.substitute(locals()))


def show_monthly_review_page(data_file=None):
    if not config.is_initialized():
        return
    global APPLICATION_URL, GRAPHICS_URL
    application_url = APPLICATION_URL
    graphics_url = GRAPHICS_URL

    web.header("Content-Type", "text/html; charset=utf-8")

    if data_file is None:
        data_file = LockablePickleFile(DATA_FILE_NAME)
        data_file.read(default=TaskList())

    unfinished_task, this_month, other_month = data_file.content.classified_task("monthly_review")
    memo = data_file.content.memo
    free_task = data_file.content.free_task
    someday_task = data_file.content.someday_task
    template = tempita.HTMLTemplate.from_filename("monthly.html")
    print adv(template.substitute(locals()))


def show_all_page(data_file=None):
    if not config.is_initialized():
        return
    global APPLICATION_URL, GRAPHICS_URL
    application_url = APPLICATION_URL
    graphics_url = GRAPHICS_URL

    web.header("Content-Type", "text/html; charset=utf-8")

    if data_file is None:
        data_file = LockablePickleFile(DATA_FILE_NAME)
        data_file.read(default=TaskList())
    monthly_list = data_file.content.monthly_task()
    all = []
    for key in sorted(monthly_list.keys()):
        date_label = "%d/%d" % (key / 100, key % 100)
        all.append([date_label, key, monthly_list[key]])
    template = tempita.HTMLTemplate.from_filename("all.html")
    print adv(template.substitute(locals()))


def post_new_task():
    data_file = LockablePickleFile(DATA_FILE_NAME)
    data_file.lock_and_read(default=TaskList())

    form = web.input()
    result = data_file.content.add(form.new_todo)
    data_file.write()
    return data_file


class index:
    def GET(self):
        show_today_page()

    def POST(self):
        if not config.is_initialized_at_post():
            return
        data_file = post_new_task()
        show_today_page(data_file)


class daily:
    def GET(self):
        show_daily_review_page()

    def POST(self):
        if not config.is_initialized_at_post():
            return
        data_file = post_new_task()
        show_daily_review_page(data_file)


class weekly:
    def GET(self):
        show_weekly_review_page()

    def POST(self):
        if not config.is_initialized_at_post():
            return
        data_file = post_new_task()
        show_weekly_review_page(data_file)


class monthly:
    def GET(self):
        show_monthly_review_page()

    def POST(self):
        if not config.is_initialized_at_post():
            return
        data_file = post_new_task()
        show_monthly_review_page(data_file)


class all:
    def GET(self):
        show_all_page()

    def POST(self):
        if not config.is_initialized_at_post():
            return
        data_file = post_new_task()
        show_all_page(data_file)


class config:
    def GET(self):
        config.show_page()

    def POST(self):
        config.save()

    @staticmethod
    def is_initialized():
        if not os.path.exists("config.txt"):
            config.show_page()
            return False
        config.read()
        return True

    @staticmethod
    def is_initialized_at_post():
        if not os.path.exists("config.txt"):
            config.save()
            return False
        config.read()
        return True

    @staticmethod
    def read():
        global APPLICATION_URL
        global GRAPHICS_URL
        global DATA_FILE_NAME
        config_file = LockablePickleFile("config.txt")
        config_file.read()
        APPLICATION_URL = config_file.content["application_url"]
        GRAPHICS_URL = config_file.content["graphics_url"]
        DATA_FILE_NAME = config_file.content["data_file_name"]
        config_file.teardown()

    @staticmethod
    def save():
        config_file = LockablePickleFile("config.txt")
        config_file.lock_and_read({})
        form = web.input()
        config_file.content["application_url"] = form.application_url.rstrip("/")
        config_file.content["graphics_url"] = form.graphics_url.rstrip("/")
        config_file.content["data_file_name"] = form.data_file_name
        config_file.write()
        show_today_page()

    @staticmethod
    def show_page():
        web.header("Content-Type", "text/html; charset=utf-8")

        is_initialized = os.path.exists("config.txt")

        data_file = LockablePickleFile("config.txt")
        data_file.read(default={"application_url":"http://",
                                "graphics_url":"http://",
                                "data_file_name":"data.txt"})

        application_url = data_file.content["application_url"]
        graphics_url = data_file.content["graphics_url"]
        data_file_name = data_file.content["data_file_name"]

        demo = is_demo()

        template = tempita.HTMLTemplate.from_filename("config.html")
        print adv(template.substitute(locals()))


class import_json:
    def GET(self):
        if not config.is_initialized():
            return
        web.header("Content-Type", "text/html; charset=utf-8")

        global APPLICATION_URL
        application_url = APPLICATION_URL
        data_file = LockablePickleFile(DATA_FILE_NAME)
        data_file.read(default=TaskList())
        json = data_file.content.dump_json().split("\n")
        data_file.teardown()

        demo = is_demo()

        template = tempita.HTMLTemplate.from_filename("import.html")
        print adv(template.substitute(locals()))

    def POST(self):
        if not config.is_initialized_at_post():
            return
        data_file = LockablePickleFile(DATA_FILE_NAME)
        data_file.lock_and_read(default=TaskList())
        form = web.input()
        data_file.content.import_json(form.import_data)
        data_file.write()
        show_today_page(data_file)


class excel_io:
    def GET(self):
        if not config.is_initialized():
            return
        web.header("Content-Type", "text/html; charset=utf-8")

        global APPLICATION_URL
        application_url = APPLICATION_URL
        data_file = LockablePickleFile(DATA_FILE_NAME)
        data_file.read(default=TaskList())
        json = data_file.content.dump_json().split("\n")
        data_file.teardown()

        template = tempita.HTMLTemplate.from_filename("import.html")
        print adv(template.substitute(locals()))

    def POST(self):
        if not config.is_initialized_at_post():
            return
        data_file = LockablePickleFile(DATA_FILE_NAME)
        data_file.lock_and_read(default=TaskList())
        form = web.input()
        data_file.content.import_json(form.import_data)
        data_file.write()
        show_today_page(data_file)

class done(object):
    def GET(self, type, finished_item_index):
        if not config.is_initialized():
            return
        data_file = LockablePickleFile(DATA_FILE_NAME)
        data_file.lock_and_read(default=TaskList())
        finished_item = data_file.content.find(type, finished_item_index)
        if finished_item is not None:
            finished_item.done()
            data_file.write()
        else:
            data_file.teardown()
        show_page(type, data_file)


class delete(object):
    def GET(self, type, delete_item_index):
        if not config.is_initialized():
            return
        data_file = LockablePickleFile(DATA_FILE_NAME)
        data_file.lock_and_read(default=TaskList())
        delete_item = data_file.content.find(type, delete_item_index)
        if delete_item is not None:
            data_file.content.delete(delete_item)
            data_file.write()
        else:
            data_file.teardown()
        show_page(type, data_file)


class repeat(object):
    def GET(self, type, copy_item_index):
        if not config.is_initialized():
            return
        data_file = LockablePickleFile(DATA_FILE_NAME)
        data_file.lock_and_read(default=TaskList())
        copy_item = data_file.content.find(type, copy_item_index)
        if copy_item is not None:
            data_file.content.copy(copy_item, 7)
            data_file.write()
        else:
            data_file.teardown()
        show_page(type, data_file)


class move(object):
    def GET(self, term, type, move_item_index):
        if not config.is_initialized():
            return
        data_file = LockablePickleFile(DATA_FILE_NAME)
        data_file.lock_and_read(default=TaskList())
        move_item = data_file.content.find(type, move_item_index)
        if move_item is not None:
            data_file.content.move(move_item, term)
            data_file.write()
        else:
            data_file.teardown()
        show_page(type, data_file)


class do_today(object):
    def GET(self, type, move_item_index):
        if not config.is_initialized():
            return
        data_file = LockablePickleFile(DATA_FILE_NAME)
        data_file.lock_and_read(default=TaskList())
        move_item = data_file.content.find(type, move_item_index)
        if move_item is not None:
            data_file.content.delete(move_item)
            time = DateCalculator.get_today()
            new_item = "%d%s" % ((time / 10000) % 10000,
                                 move_item._title)
            daytask = move_item.to(DayTask)
            daytask._target_time = DateCalculator.get_now()
            data_file.content.add_object(daytask)
            data_file.write()
        else:
            data_file.teardown()
        show_page(type, data_file)


if __name__ == "__main__":
    web.webapi.internalerror = web.debugerror
    web.run(urls, globals())
