#! d:/Python26/python.exe

import os
import re
import web
import copy
import time
import heapq
import bisect
import hashlib
import tempita
import calendar
import datetime
import itertools

from web_helper import *

try:
    import json
except ImportError:
    import simplejson as json

# =====================================
# Model
# =====================================

class TimeComparable(object):
    def __cmp__(self, other):
        return cmp(self.time(), other.time())

    def time():
        raise NotImplementedError()


class DateFragment(TimeComparable):
    def __init__(self, time):
        self._time = time

    def time(self):
        return self._time


class DateCalculator(object):
    _cache = {}

    @classmethod
    def is_past(cls, other):
        return other < cls.today_start()

    @classmethod
    def today_start(cls):
        result = cls._cache.get("today_start")
        if result is None:
            result = int(time.strftime("%Y%m%d0000"))-1
            cls._cache["today_start"] = result
        return result

    @classmethod
    def today_end(cls):
        result = cls._cache.get("today_end")
        if result is None:
            today_end = datetime.date.today()
            today_end += datetime.date.resolution
            result =  cls.to_int(today_end) - 1
            cls._cache["today_end"] = result
        return result

    @classmethod
    def week_start(cls):
        result = cls._cache.get("week_start")
        if result is None:
            week_start = datetime.date.today()
            while week_start.weekday() != calendar.MONDAY:
                week_start -= datetime.date.resolution
            result = cls.to_int(week_start) - 1
            cls._cache["week_start"] = result
        return result

    @classmethod
    def week_end(cls):
        result = cls._cache.get("week_end")
        if result is None:
            result = cls.shift_days(cls.week_start() + 1, 7) - 1
            cls._cache["week_end"] = result
        return result

    @classmethod
    def month_start(cls):
        result = cls._cache.get("month_start")
        if result is None:
            result = int(time.strftime("%Y%m010000"))-1
            cls._cache["month_start"] = result
        return result

    @classmethod
    def month_end(cls):
        result = cls._cache.get("month_end")
        if result is None:
            today = datetime.date.today()
            month_end = datetime.date.today()
            while today.month == month_end.month:
                month_end += datetime.date.resolution
            result = cls.to_int(month_end) - 1
            cls._cache["month_end"] = result
        return result

    @classmethod
    def get_today(cls):
        return cls.today_start() + 1

    @classmethod
    def get_today_as_str(cls):
        result = cls._cache.get("today_as_str")
        if result is None:
           result = time.strftime("%m%d")
           cls._cache["today_as_str"] = result
        return result

    @classmethod
    def get_now(cls):
        result = cls._cache.get("now")
        if result is None:
           result = int(time.strftime("%Y%m%d%H%M"))
           cls._cache["now"] = result
        return result

    @classmethod
    def to_int(cls, date):
        time = (date.year * 10000 + date.month * 100 + date.day) * 10000
        if isinstance(date, datetime.datetime):
            time += date.hour * 100 + date.minute
        return time

    @classmethod
    def shift_days(cls, input_time, delta=0):
        time_struct = time.strptime(str(input_time), "%Y%m%d%H%M")
        result = datetime.datetime(*time_struct[:5])
        result += datetime.timedelta(days=delta)
        return cls.to_int(result)

    @classmethod
    def shift_to_next_week(cls, input_time):
        time_struct = time.strptime(str(input_time), "%Y%m%d%H%M")
        next_week = datetime.datetime(*time_struct[:5])
        if next_week.weekday() == calendar.MONDAY:
            next_week += datetime.timedelta(days=7)
        else:
            while next_week.weekday() != calendar.MONDAY:
                next_week += datetime.date.resolution
        return cls.to_int(next_week)

    @classmethod
    def shift_to_next_month(cls, input_time):
        time_struct = time.strptime(str(input_time), "%Y%m%d%H%M")
        target_month = (time_struct[1] + 1) % 12
        result = datetime.datetime(*time_struct[:5])
        while target_month != result.month:
              result += datetime.date.resolution
        return cls.to_int(result)


class TaskList(object):
    def __init__(self):
        self.scheduled_task = []
        self.free_task = []
        self.someday_task = []
        self.memo = []

    def classified_task(self, target = None):
        before_term = []
        trash = []
        month_1st_half = []
        month_2nd_half = []
        month = []
        week_1st_half = []
        week_2nd_half = []
        week = []
        today = []

        if target == "today":
            breakpoints = [DateFragment(DateCalculator.today_start()),
                           DateFragment(DateCalculator.today_end())]
            group = [before_term, today, trash]
        elif target == "daily_review":
            breakpoints = [DateFragment(DateCalculator.week_start()),
                           DateFragment(DateCalculator.today_start()),
                           DateFragment(DateCalculator.today_end()),
                           DateFragment(DateCalculator.week_end())]
            group = [before_term, week_1st_half, today, week_2nd_half, trash]
        elif target == "weekly_review":
            breakpoints = [DateFragment(DateCalculator.month_start()),
                           DateFragment(DateCalculator.week_start()),
                           DateFragment(DateCalculator.week_end()),
                           DateFragment(DateCalculator.month_end())]
            group = [before_term, month_1st_half, week, month_2nd_half, trash]
        elif target == "monthly_review":
            breakpoints = [DateFragment(DateCalculator.month_start()),
                           DateFragment(DateCalculator.week_start()),
                           DateFragment(DateCalculator.week_end()),
                           DateFragment(DateCalculator.month_end())]
            group = [before_term, month_1st_half, week, month_2nd_half, trash]
        else:
            breakpoints = [DateFragment(DateCalculator.month_start()),
                           DateFragment(DateCalculator.week_start()),
                           DateFragment(DateCalculator.today_start()),
                           DateFragment(DateCalculator.today_end()),
                           DateFragment(DateCalculator.week_end()),
                           DateFragment(DateCalculator.month_end())]
            group = [trash, month, week, today,
                     week, month, trash]

        for task in sorted(self.scheduled_task):
            index = bisect.bisect(breakpoints, task)
            group[index].append(task)

        if target == "today":
            unfinished = [task for task in before_term if not task.is_done()]
            return unfinished, today
        elif target == "daily_review":
            unfinished = [task for task in before_term if not task.is_done()]
            for task in week_1st_half:
                if task.is_done():
                    week.append(task)
                else:
                    unfinished.append(task)
            return unfinished, week + week_2nd_half, today
        elif target == "weekly_review":
            dprint("weekly_review filter done")
            unfinished = [task for task in before_term if not task.is_done()]
            for task in month_1st_half:
                if task.is_done():
                    month.append(task)
                else:
                    unfinished.append(task)
            return unfinished, month + month_2nd_half, week
        else:
            return month, week, today

    def monthly_task(self, source=None):
        if source is None:
            source = self.scheduled_task
        monthly_list = {}
        for task in sorted(source, key=lambda t: t.target_time()):
            monthly_list.setdefault(task.month_key(), []).append(task)
        return monthly_list

    def find(self, type, item_index):
        type = type.split("_")[-1]
        if type == "unfinished":
            unfinished, today = self.classified_task("today")
            return unfinished[int(item_index)]
        elif type == "today":
            unfinished, today = self.classified_task("today")
            return today[int(item_index)]
        elif type == "week":
            month, week, today = self.classified_task()
            return week[int(item_index)]
        elif type == "month" and "_" not in item_index:
            month, week, today = self.classified_task()
            return month[int(item_index)]
        elif type == "someday":
            return self.someday_task[int(item_index)]
        elif type == "free":
            return self.free_task[int(item_index)]
        elif type == "memo":
            return self.memo[int(item_index)]
        elif "_" in item_index:
            date_key, index = map(int, item_index.split("_"))
            return self.monthly_task()[date_key][index]

    def add(self, title):
        task_classes = (TimedTask, FreeTask, SomedayTask,
                        DayTask, TaskWithDeadline, Memo)
        for task_class in task_classes:
            result = task_class.parse_from_input_str(title)
            if result:
                self.add_object(result)
                break
        return result

    def add_object(self, new_task):
        task_list = {TimedTask: self.scheduled_task,
                     FreeTask: self.free_task,
                     SomedayTask: self.someday_task,
                     DayTask: self.scheduled_task,
                     Memo: self.memo,
                     TaskWithDeadline: self.scheduled_task}
        bisect.insort(task_list[new_task.__class__], new_task)

    def delete(self, delete_item):
        class_list = [((DayTask, TimedTask), self.scheduled_task),
                      (Memo, self.memo),
                      (SomedayTask, self.someday_task),
                      (FreeTask, self.free_task)]
        is_delete = False
        for class_type, storage in class_list:
            if is_delete:
                break
            if isinstance(delete_item, class_type):
                for i, task in enumerate(storage):
                    if task is delete_item:
                        del storage[i]
                        is_delete = True
                        break

    def copy(self, copy_item, day):
        if isinstance(copy_item, Memo):
            return
        copied_item = copy.copy(copy_item)
        while copied_item.is_done():
            copied_item.done()
        copied_item._target_time = DateCalculator.shift_days(copied_item._target_time, day)
        bisect.insort(self.scheduled_task, copied_item)

    def move(self, move_item, term):
        if isinstance(move_item, Memo):
            return
        elif term == "tomorrow":
            move_item._target_time = DateCalculator.shift_days(move_item._target_time, 1)
        elif term == "next_week":
            move_item._target_time = DateCalculator.shift_to_next_week(move_item._target_time)
        elif term == "next_month":
            move_item._target_time = DateCalculator.shift_to_next_month(move_item._target_time)
        elif term == "someday":
            self.delete(move_item)
            self.someday_task.append(move_item.to(SomedayTask))
        self.scheduled_task.sort()

    def dump_json(self):
        scheduled_task = [task.dump_json() for task in self.scheduled_task]
        free_task = [task.dump_json() for task in self.free_task]
        someday_task = [task.dump_json() for task in self.someday_task]
        memo = [memo.dump_json() for memo in self.memo]
        result = ",\n  ".join(scheduled_task + free_task + someday_task + memo)
        return "[\n  %s\n]" % result

    def import_json(self, json_raw_data):
        self.scheduled_task = []
        self.free_task = []
        self.someday_task = []
        self.memo = []

        task_list = {"timed":(TimedTask, self.scheduled_task),
                     "free":(FreeTask, self.free_task),
                     "someday":(SomedayTask, self.someday_task),
                     "memo":(Memo, self.memo),
                     "day":(DayTask, self.scheduled_task),
                     "deadline":(TaskWithDeadline, self.scheduled_task)}

        for raw_data in json.loads(json_raw_data):
            task_class, storage = task_list[raw_data[0]]
            dprint("importing... %s" % raw_data[0])
            result = task_class.from_json(raw_data)
            if result:
                bisect.insort(storage, result)


class Entity(TimeComparable):
    def __init__(self, title, time=None, year=None):
        self._create_time = DateCalculator.get_now()
        self._target_time = None
        self._start_time = None
        self._finished_time = None
        if time or year:
            self._set_time(time, year)
        self._set_title(title)

    def _set_title(self, title):
        title = title.strip()
        if title.startswith("'"):
            self._title = title[1:]
            if not self._finished_time:
                self._finished_time = True
        else:
            self._title = title

    def done(self):
        if self.is_done():
            self._finished_time = None
            if self._title.startswith("'"):
                self._title = self._title[1:]
        else:
            self._finished_time = DateCalculator.get_now()

    def is_done(self):
        return self._finished_time is not None

    def is_today(self):
        return ((DateCalculator.today_start() < self.time())
                and (self.time() < DateCalculator.today_end()))

    def css_class(self):
        if self.is_done():
            return "done"
        return "active"

    def _to_datetime_format(self, target=False):
        if target:
            base_time = self.target_time()
        else:
            base_time = self.time()
        date = (base_time / 10000) % 10000
        time = base_time % 10000
        return "%d/%d %d:%02d" % (date/100, date%100, time/100, time%100)

    def _to_date_format(self, target=False):
        if target:
            base_time = self.target_time()
        else:
            base_time = self.time()
        date = (base_time / 10000) % 10000
        return "%d/%d" % (date/100, date%100)

    @classmethod
    def parse_from_input_str(cls, input_str):
        raise NotImplementedError()

    def time(self):
        raise NotImplementedError()

    def target_time(self):
        return self._target_time

    def dump_json(self):
        times = {}
        times["create"] = self._create_time
        times["target"] = self._target_time
        times["start"] = self._start_time
        times["finish"] = self._finished_time

        return  json.dumps([self.json_type, times, self._title])

    @classmethod
    def from_json(cls, json_object):
        new_task = cls(json_object[2])
        new_task._create_time = json_object[1].get("create")
        new_task._target_time = json_object[1].get("target")
        new_task._start_time = json_object[1].get("start")
        new_task._finished_time = json_object[1].get("finish")
        return new_task

    def to(self, cls):
        new_task = cls(self._title)
        if self._create_time:
            new_task._create_time = self._create_time
        else:
            new_task._create_time = DateCalculator.get_now()
        if isinstance(new_task, (DayTask, TimedTask, TaskWithDeadline)):
            if self._target_time:
                new_task._target_time = self._target_time
            else:
                new_task._target_time = DateCalculator.get_today()
        new_task._start_time = self._start_time
        new_task._finished_time = self._finished_time
        return new_task


class Memo(Entity):
    input_pattern = re.compile("(.*)")
    json_type = "memo"

    def __init__(self, title, time=None):
        super(Memo, self).__init__(title, time)

    def _set_time(self, time, year):
        if time is None:
            time = DateCalculator.to_int(datetime.datetime.now())
        self._create_time = int(time)

    def time(self):
        return self._create_time

    @classmethod
    def parse_from_input_str(cls, input_str):
        match = cls.input_pattern.match(input_str)
        if match:
            return cls(match.group(1))

    def to_html(self):
        return "%s: %s" % (self._to_datetime_format(), self._title)


class FreeTask(Memo):
    input_pattern = re.compile(">(.*)")
    json_type = "free"

    def __init__(self, title, time=None):
        super(FreeTask, self).__init__(title, time)

    def to_html(self):
        return "%s (registered at %s)" % (self._title, self._to_datetime_format())


class SomedayTask(Memo):
    input_pattern = re.compile("#(.*)")
    json_type = "someday"

    def __init__(self, title, time=None):
        super(SomedayTask, self).__init__(title, time)

    def css_class(self):
        if self.is_done():
            return "done"
        return "active"

    def to_html(self):
        return "%s: %s" % (self._to_datetime_format(), self._title)


class DayTask(Entity):
    input_pattern = re.compile("(?:(\\d{4})|@)(.*)")
    json_type = "day"

    def __init__(self, title, date=None, year=None):
        super(DayTask, self).__init__(title, date, year)

    @classmethod
    def parse_from_input_str(cls, input_str):
        match = cls.input_pattern.match(input_str)
        if not match:
            return
        date = match.group(1)
        if date is None:
            date = DateCalculator.get_today_as_str()
        return cls(match.group(2), date)

    def _set_time(self, mmdd, year):
        year = year or time.localtime()[0]
        self._target_time = int(year) * 10000 * 10000 + int(mmdd) * 10000
        if DateCalculator.is_past(self._target_time):
            self._target_time += 1 * 10000 * 10000

    def time(self):
        return self._target_time

    def css_class(self):
        if self.is_done():
            if DateCalculator.is_past(self.time()):
                return "old_and_done"
            return "done"
        else:
            if DateCalculator.is_past(self.time()):
                return "old"
            return "active"

    def date(self):
        return self._to_date_format()

    def to_html(self):
        return "%s: %s" % (self.date(), self._title)

    def month_key(self):
        return self.time() / 10000 / 100

    def is_timed_task(self):
        return False


class TimedTask(DayTask):
    input_pattern = re.compile("(\\d{8})(.*)")
    json_type = "timed"

    def __init__(self, title, date_time=None, year=None):
        super(TimedTask, self).__init__(title, date_time, year)

    def _set_time(self, time_str, year):
        year = year or time.localtime()[0]
        self._target_time = int(year) * 10000 * 10000 + int(time_str)
        if DateCalculator.is_past(self._target_time):
            self._target_time += 1 * 10000 * 10000

    def date(self):
        return self._to_datetime_format()

    def is_timed_task(self):
        return True


class TaskWithDeadline(DayTask):
    input_pattern = re.compile("\\.+(\\d{4})(.*)")
    json_type = "deadline"

    def __init__(self, title, mmdd=None, year=None):
        super(TaskWithDeadline, self).__init__(title, mmdd, year)

    def _set_time(self, mmdd, year):
        year = year or time.localtime()[0]
        self._target_time = int(year) * 10000 * 10000 + int(mmdd) * 10000
        if DateCalculator.is_past(self._target_time):
            self._target_time += 1 * 10000 * 10000

    def date(self):
        return self._to_date_format(target=True)

    def is_timed_task(self):
        return True

    def time(self):
        delta = float(self.target_time())
        while delta > 1:
            delta = delta * 0.1
        return DateCalculator.today_end() - 1 + delta

    def css_class(self):
        if self.is_done():
            if DateCalculator.is_past(self.target_time()):
                return "old_and_done"
            return "done"
        else:
            if DateCalculator.is_past(self.target_time()):
                return "old"
            return "active"

    def to_html(self):
        return "%s (Limit: %s)" % (self._title, self.date())

