import os, pickle

# =====================================
# Helper Function
# =====================================

def dprint(a):
    import pprint
    log = file("log.txt", "a")
    pprint.pprint(a, log)


# =====================================
# Model
# =====================================

class LockablePickleFile(object):
    def __init__(self, filename):
        self.file_obj = None
        self.filename = filename
        self.content = None
        self._oldcontent = None

    def read(self, default=None):
        if os.path.exists(self.filename):
            try:
                file_content = file(self.filename).read()
                self.content = pickle.loads(file_content)
                self._oldcontent = pickle.loads(file_content)
            except:
                self.content = default
                self._oldcontent = default
        else:
            self.content = default
            self._oldcontent = default

    def lock_and_read(self, default=None):
        self.read(default)
        self.file_obj = file(self.filename, "w")
        self._lock(self.file_obj, LockablePickleFile.LOCK_EX)

    def write(self):
        if self.file_obj is None:
            raise RuntimeError("This file is not locked")
        pickle.dump(self.content, self.file_obj)
        self._unlock(self.file_obj)
        self.file_obj = None

    def teardown(self):
        if self.file_obj is None:
            return
        self.content = self._oldcontent
        self.write()

    def rollback(self):
        if self.file_obj is not None:
            self._unlock(self.file_obj)
            self.file_obj = None
        self.content = self._oldcontent

    def __del__(self):
        self.teardown()

    if os.name == 'nt':
        from win32con import LOCKFILE_EXCLUSIVE_LOCK as LOCK_EX
        LOCK_SH = 0
        from win32con import LOCKFILE_FAIL_IMMEDIATELY as LOCK_NB

        @staticmethod
        def _lock(file, flags):
            import win32file, pywintypes
            overlapped = pywintypes.OVERLAPPED()
            hfile = win32file._get_osfhandle(file.fileno())
            win32file.LockFileEx(hfile, flags, 0, -0x10000, overlapped)

        @staticmethod
        def _unlock(file):
            import win32file, pywintypes
            overlapped = pywintypes.OVERLAPPED()
            hfile = win32file._get_osfhandle(file.fileno())
            win32file.UnlockFileEx(hfile, 0, -0x10000, overlapped)

    elif os.name == 'posix':
        from fcntl import LOCK_EX, LOCK_SH, LOCK_NB

        @staticmethod
        def _lock(file, flags):
            import fcntl
            fcntl.flock(file.fileno(), flags)

        @staticmethod
        def _unlock(file):
            import fcntl
            fcntl.flock(file.fileno(), fcntl.LOCK_UN)
    else:
        raise RuntimeError()



